#define DIRLIST_H

/****************************************************************

    Copyright (c) 1996-1997 Fred Pesch

    This program is free software; you can redistribute it 
    and/or modify it under the terms of the GNU General 
    Public License as published by the Free Software 
    Foundation (version 2 of the License).

    This program is distributed in the hope that it will 
    be useful, but WITHOUT ANY WARRANTY; without even the 
    implied warranty of MERCHANTABILITY or FITNESS FOR A 
    PARTICULAR PURPOSE.  See the GNU General Public License 
    for more details.

    For a copy of the GNU General Public License, write to the 
    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
    MA 02139, USA.

****************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Xm/Xm.h>

#include "error_handler.h"
#include "config.h"

extern uid_t UserID;
extern gid_t GroupID;

typedef enum { NONE = 0, READ = 1, WRITE = 2, EXECUTE = 4 } SECURITY_FLAGS;

typedef struct {
  SECURITY_FLAGS user;
  SECURITY_FLAGS group;
  SECURITY_FLAGS others;

  int opened;
  int readable;

  uid_t UID;
  gid_t GID;

  time_t changetime;
  time_t modiftime;
  time_t accesstime;

  off_t size;

  char *pathname;
  char *name;
} TDirEntry;

  
struct DirNode {
  TDirEntry dir;

  struct DirNode *parent;

  int nb_children;
  struct DirNode **child_list;
};

typedef struct DirNode TDirNode;

struct SortDirEntry {
  struct stat statbuf;
  char *name;
};

typedef struct SortDirEntry TSortDirEntry;


/********************************
 * CreateRootNode               *
 ********************************
 * Create the first entry       *
 ********************************/

TDirNode *CreateRootNode(void)
{
  TDirNode *node;
  int result;

  node = (TDirNode *) XtMalloc(sizeof(TDirNode));
  if(node==NULL) FatalError("dirlist.c","CreateRootNode","Unable to create new node (XtMalloc fails)");

  result = chdir("/");
  if(result==-1) FatalError("dirlist.c","CreateRootNode","Cannot chdir to root dir");

  node->nb_children=0;
  node->child_list=NULL;

  node->parent=NULL;

  node->dir.user = (READ | WRITE | EXECUTE);
  node->dir.group = (READ | WRITE | EXECUTE);
  node->dir.others =  (READ | WRITE | EXECUTE);

  node->dir.name=(char *) XtMalloc(sizeof(char)*2);
  if(node->dir.name==NULL) FatalError("dirlist.c","CreateRootNode","Unable to create new node (XtMalloc fails)");
  strcpy(node->dir.name, "/");

  node->dir.pathname=(char *) XtMalloc(sizeof(char)*1);
  if(node->dir.pathname==NULL) FatalError("dirlist.c","CreateRootNode","Unable to create new node (XtMalloc fails)");
  strcpy(node->dir.pathname,"");

  node->dir.UID=000;
  node->dir.GID=000;

  node->dir.modiftime=0;
  node->dir.changetime=0;
  node->dir.accesstime=0;

  node->dir.size=0;

  node->dir.readable=1;
  node->dir.opened=0;

  return(node);
}
  

/********************************
 * CreateSubDirNode             *
 ********************************
 * Create a new node for a      *
 * sub-directory                *
 ********************************/

TDirNode *CreateSubdirNode(struct stat *statbuf, TDirNode *parent, char *name)
{
  TDirNode *node;
  umode_t mode;

  node = (TDirNode *) XtMalloc(sizeof(TDirNode));
  if(node==NULL) FatalError("dirlist.c","CreateSubdirNode","Unable to create new node (XtMalloc fails)");

  mode = statbuf->st_mode;

  /* Get Parent */
  node->parent = parent;

  /* Get Name */
  node->dir.name=(char *) XtMalloc(sizeof(char)*(strlen(name)+1));
  if(node->dir.name==NULL) FatalError("dirlist.c","CreateSubdirNode","Unable to create new node (XtMalloc fails)"); 
  strcpy(node->dir.name, name);

  /* Get PathName */
  node->dir.pathname=(char *) XtMalloc(sizeof(char)*(strlen(name)+1+strlen(parent->dir.pathname)+1+1));
  if(node->dir.pathname==NULL) FatalError("dirlist.c","CreateSubdirNode","Unable to create new node (XtMalloc fails)");
  strcpy(node->dir.pathname, parent->dir.pathname);
  strcat(node->dir.pathname, "/");
  strcat(node->dir.pathname, name);

  /* Get UID */
  node->dir.UID = statbuf->st_uid;
  
  /* Get GID */
  node->dir.GID = statbuf->st_gid;

  /* Get Size */
  node->dir.size = statbuf->st_size;

  /* Get Dates */
  node->dir.changetime = statbuf->st_ctime;
  node->dir.modiftime = statbuf->st_mtime;
  node->dir.accesstime = statbuf->st_atime;

  

  /* Get User protections */
  node->dir.user=NONE;

  if(mode & S_IRUSR) node->dir.user |= READ;
  if(mode & S_IWUSR) node->dir.user |= WRITE;
  if(mode & S_IXUSR) node->dir.user |= EXECUTE;

  /* Get Group protections */
  node->dir.group=NONE;

  if(mode & S_IRGRP) node->dir.group |= READ;
  if(mode & S_IWGRP) node->dir.group |= WRITE;
  if(mode & S_IXGRP) node->dir.group |= EXECUTE;

  /* Get Others protections */
  node->dir.others=NONE;

  if(mode & S_IROTH) node->dir.others |= READ;
  if(mode & S_IWOTH) node->dir.others |= WRITE;
  if(mode & S_IXOTH) node->dir.others |= EXECUTE;

  /* default values */

  if(((node->dir.user & READ) && (node->dir.UID==UserID)) || 
     ((node->dir.group & READ) &&(node->dir.GID==GroupID)) ||
     (node->dir.others & READ)) 
    node->dir.readable=1; 
  else node->dir.readable=0;
  
  node->dir.opened=0;

  node->nb_children=0;
  node->child_list=NULL;

  return(node);
}


int CompareDirName(const TSortDirEntry *buf1, const TSortDirEntry *buf2) 
{
  return (strcmp(buf1->name, buf2->name));
}

/********************************
 * MakeDirOneLevel              *
 ********************************
 * finds all the subdirs of     *
 * the directory argument       *
 ********************************/

void MakeDirOneLevel(TDirNode *actualnode)
{
  int result;
  int i;
  struct dirent *entry;
  struct stat statbuf;
  DIR *dir;

  TSortDirEntry *sortlist=NULL;

  actualnode->dir.opened=1;

  /* Go to the directory */
  if(actualnode->parent==NULL) result = chdir("/");
  else result = chdir(actualnode->dir.pathname);
  if(result==-1) {
    NonFatalError("dirlist.c","MakeDirOneLevel","Cannot chdir to actualnode pathname");
    return;
  }

  /* open the directory */
  dir=NULL;
  dir = opendir(".");
  if(dir==NULL) {
    NonFatalError("dirlist.c","MakeDirOneLevel","Cannot open directory");
    return;
  };

  /* read subfiles */
  actualnode->nb_children=0;
  do {
    entry = readdir(dir);
    if(entry!=NULL) {
      result = stat(entry->d_name, &statbuf);
      /*      if(result==-1) FatalError("dirlist.c","MakeDirOneLevel","Cannot stat directory entry (stat failed)");*/

      /*      if(result==-1) fprintf(stderr, "\nWarning : cannot stat : %s\n",entry->d_name);*/

      /* if read file is a directory */
      if(result!=-1 && S_ISDIR(statbuf.st_mode) && (strcmp(entry->d_name, ".")!=0) && (strcmp(entry->d_name, "..")!=0)) {
	
	/*	fprintf(stderr, "%s\n", entry->d_name); */

	actualnode->nb_children++;
	
	if(actualnode->nb_children==1) 
	  sortlist= (TSortDirEntry *) XtMalloc(sizeof(TSortDirEntry)*actualnode->nb_children);
	else
	  sortlist= (TSortDirEntry *) XtRealloc( (char *) sortlist, sizeof(TSortDirEntry)*actualnode->nb_children);

	memcpy(&sortlist[actualnode->nb_children-1].statbuf,&statbuf,sizeof(struct stat));
	sortlist[actualnode->nb_children-1].name= XtMalloc(sizeof(char)*(strlen(entry->d_name)+1));
	strcpy(sortlist[actualnode->nb_children-1].name, entry->d_name);

      }
    }
  } while(entry!=NULL);

  if(actualnode->nb_children>0) {
    actualnode->child_list = (TDirNode **) XtRealloc((char *) actualnode->child_list, sizeof(TDirNode *)*actualnode->nb_children);

    qsort(sortlist, actualnode->nb_children, sizeof(TSortDirEntry), (const void *) CompareDirName);

    for(i=0;i<actualnode->nb_children;i++) {
      actualnode->child_list[i] = CreateSubdirNode(&sortlist[i].statbuf, actualnode, sortlist[i].name);
      XtFree( sortlist[i].name);
    }

    XtFree( (char *) sortlist);
  }

  closedir(dir);
}

/********************************
 * DeleteSubdirNode             *
 ********************************
 * delete recurively all        *
 * subdirs of current node      *
 ********************************/

void DeleteSubdirNode(TDirNode *parent)
{
  int i;

  for(i=0;i<parent->nb_children;i++) {
    if(parent->child_list[i]!=NULL) DeleteSubdirNode(parent->child_list[i]);
    XtFree(parent->child_list[i]->dir.name);
    XtFree(parent->child_list[i]->dir.pathname);
    XtFree((char *)parent->child_list[i]);
    parent->child_list[i]=NULL;
  };
  XtFree((char *)parent->child_list);
  parent->child_list=NULL;
  parent->nb_children=0;
  parent->dir.opened=0;
}

/********************************
 * DeleteRootDir                *
 ********************************
 * Delete first entry           *
 ********************************/      

void DeleteRootDir(TDirNode *node)
{
  DeleteSubdirNode(node);
  XtFree(node->dir.name);
  XtFree(node->dir.pathname);
  XtFree((char *)node);
}
    



  
  



